
// File: <graphics_framework.cpp>
// Copyright (c) 1997-2007, Reachin Technologies AB.
// All rights reserved.
// The copyright to the computer program herein is the property of Reachin
// Technologies AB, Sweden. The program may be used and/or copied only with the
// written permission of Reachin Technologies AB or in accordance with the terms
// stipulated in the agreement under which the program has been supplied.

#include "graphics_framework.h"
#include "libASE.h"
#include <math.h>

#pragma comment(lib,"HaptX/libASE.lib")

bool fAlmostEqual(float A, float B, int maxUlps)
{
    if (!(maxUlps > 0 && maxUlps < 4 * 1024 * 1024))
		return false;

    int aInt = *(int*)&A;
	if (aInt < 0) aInt = 0x80000000 - aInt;

    int bInt = *(int*)&B;
    if (bInt < 0) bInt = 0x80000000 - bInt;

    int intDiff = abs(aInt - bInt);

    return (intDiff <= maxUlps);
}

// ###############################################################################################################

GraphicsRenderObject::GraphicsRenderObject()
{
	m_vertices = 0;
	m_faces = 0;
	m_numFaces = 0;
	m_numVertices = 0;
	m_colormap = 0;
	m_tissueType = 0;
}

GraphicsRenderObject::~GraphicsRenderObject()
{
}

void GraphicsRenderObject::FreeObject()
{
	if (m_vertices) delete [] m_vertices;
	if (m_faces) delete [] m_faces;
	if (m_colormap) delete m_colormap;

	m_vertices = 0;
	m_faces = 0;
	m_numFaces = 0;
	m_numVertices = 0;
	m_colormap = 0;
}

void GraphicsRenderObject::Clone(GraphicsRenderObject* obj)
{
	m_numVertices = obj->m_numVertices;
	m_numFaces = obj->m_numFaces;

	if (obj->m_vertices)
	{
		m_vertices = new vertex_t[obj->m_numVertices];
		memcpy(m_vertices,obj->m_vertices,sizeof(vertex_t) * obj->m_numVertices);
	}

	if (obj->m_faces)
	{
		m_faces = new unsigned int[obj->m_numFaces * 3];
		memcpy(m_faces,obj->m_faces,sizeof(unsigned int) * obj->m_numFaces * 3);
	}

	if (obj->m_colormap) obj->m_colormap->Clone(m_colormap);
}

void GraphicsRenderObject::ComputeNormals(bool right_handed)
{
	if (m_numFaces < 1 || !m_faces || !m_vertices) return;
	float* faceNormals = new float[m_numFaces * 3];
	int i, j, k1, k2;
	unsigned int id[3], idv, idu;
	float v[3], u[3], n[3], l;

	//  compute face normals
	for (i = 0; i < m_numFaces; i++)
	{
		id[0] = m_faces[i * 3 + 0];
		id[1] = m_faces[i * 3 + 1];
		id[2] = m_faces[i * 3 + 2];

		m_vertices[id[0]].normal[0] = m_vertices[id[0]].normal[1] = m_vertices[id[0]].normal[2] = 0.0f;
		m_vertices[id[1]].normal[0] = m_vertices[id[1]].normal[1] = m_vertices[id[1]].normal[2] = 0.0f;
		m_vertices[id[2]].normal[0] = m_vertices[id[2]].normal[1] = m_vertices[id[2]].normal[2] = 0.0f;
		
		v[0] = m_vertices[id[1]].vertex[0] - m_vertices[id[2]].vertex[0];
		v[1] = m_vertices[id[1]].vertex[1] - m_vertices[id[2]].vertex[1];
		v[2] = m_vertices[id[1]].vertex[2] - m_vertices[id[2]].vertex[2];

		u[0] = m_vertices[id[0]].vertex[0] - m_vertices[id[2]].vertex[0];
		u[1] = m_vertices[id[0]].vertex[1] - m_vertices[id[2]].vertex[1];
		u[2] = m_vertices[id[0]].vertex[2] - m_vertices[id[2]].vertex[2];

		faceNormals[i * 3 + 0] = u[1] * v[2] - u[2] * v[1];
		faceNormals[i * 3 + 1] = u[2] * v[0] - u[0] * v[2];
		faceNormals[i * 3 + 2] = u[0] * v[1] - u[1] * v[0];

		l = sqrtf(faceNormals[i * 3 + 0] * faceNormals[i * 3 + 0] + faceNormals[i * 3 + 1] * faceNormals[i * 3 + 1] + faceNormals[i * 3 + 2] * faceNormals[i * 3 + 2]);

		if (l == 0.0f) l = 0.0001f;

		faceNormals[i * 3 + 0] /= l;
		faceNormals[i * 3 + 1] /= l;
		faceNormals[i * 3 + 2] /= l;
	}

	// compute vertex normals (does not use vertex weight, just a simple and in some cases incorrect way of calculating normals)
	// this is to slow !!!!!
	// cant have this on the tetrahedral meshes !!!!!!!!!!!
	for (i = 0; i < m_numFaces; i++)
	{
		for (k1 = 0; k1 < 3; k1++)
		{
			idv = m_faces[i * 3 + k1];

			n[0] = m_vertices[idv].normal[0];
			n[1] = m_vertices[idv].normal[1];
			n[2] = m_vertices[idv].normal[2];

			if (!(n[0] == 0.0f && n[1] == 0.0f && n[2] == 0.0f))
				continue;

			for (j = 0; j < m_numFaces; j++)
			{
				for (k2 = 0; k2 < 3; k2++)
				{
					idu = m_faces[j * 3 + k2];

					if (idv == idu || (fAlmostEqual(m_vertices[idv].vertex[0],m_vertices[idu].vertex[0],10) && fAlmostEqual(m_vertices[idv].vertex[1],m_vertices[idu].vertex[1],10) && fAlmostEqual(m_vertices[idv].vertex[2],m_vertices[idu].vertex[2],10)))
					{
						n[0] += faceNormals[j * 3 + 0];
						n[1] += faceNormals[j * 3 + 1];
						n[2] += faceNormals[j * 3 + 2];
					}
				}
			}

			l = sqrtf(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);
			if (l == 0.0f) l = 0.0001f;
			n[0] /= l; n[1] /= l; n[2] /= l;

			m_vertices[idv].normal[0] = n[0];
			m_vertices[idv].normal[1] = n[1];
			m_vertices[idv].normal[2] = n[2];
		}
	}

	delete [] faceNormals;
}

// ###############################################################################################################

void IGraphicsRenderModel::ComputeNormals(bool right_handed)
{
	for (size_t i = 0; i < m_objectArray.size(); i++)
	{
		m_objectArray[i].ComputeNormals(right_handed);
	}
}

// ###############################################################################################################

image_t IGraphicsFramework::LoadImageBMP(const char* strFile)
{
	BITMAPFILEHEADER bfh;
	BITMAPINFOHEADER bih;
	image_t img;
	img.data = 0;

	FILE* fp = fopen(strFile,"rb");
	if (!fp) return img;

	fread(&bfh,sizeof(BITMAPFILEHEADER),1,fp);
	if (bfh.bfType != 0x4D42)
	{
		fclose(fp);
		return img;
	}

	fread(&bih,sizeof(BITMAPINFOHEADER),1,fp);
	if (bih.biBitCount != 24)
	{
		fclose(fp);
		return img;
	}

	img.width = (int)bih.biWidth;
	img.height = (int)bih.biHeight;
	img.data = new unsigned char[img.width * img.height * 3];

	fread(img.data,sizeof(unsigned char),img.width * img.height * 3,fp);

	// flip the data from BGR to RGB
	for (int i = 0; i < img.width; i++)
	{
		for (int j = 0; j < img.height; j++)
		{
			unsigned char t = img.data[(j * img.width + i) * 3 + 0];
			img.data[(j * img.width + i) * 3 + 0] = img.data[(j * img.width + i) * 3 + 2];
			img.data[(j * img.width + i) * 3 + 2] = t;
		}
	}

	fclose(fp);

	return img;
}

IGraphicsRenderModel* IGraphicsFramework::LoadRenderModel(const char* strFile)										// load a render model (obky ASE files are supported)
{
	IGraphicsRenderModel* m = CreateRenderModel();
	bool useTextureCoordArray;

	if (m)
	{
		ASE_Scene* scene = ASE_loadFilename((char*)strFile);
		
		if (scene)
		{
			for (int i = 0; i < scene->objectCount; i++)
			{
				GraphicsRenderObject obj;
				m->m_objectArray.push_back(obj);
				size_t id = m->m_objectArray.size() - 1;

				int material_id = scene->objs[i].material;
				if (material_id > -1 && material_id < scene->materialCount)
				{
					m->m_objectArray[id].m_colormap = LoadTexture(scene->materials[material_id].diffuseMaterialMap.image_path);
				}

				int vertex_count = scene->objs[i].mesh.vertexCount;
				if (vertex_count < scene->objs[i].mesh.textureCoordinateCount)
				{
					vertex_count = scene->objs[i].mesh.textureCoordinateCount;
					useTextureCoordArray = true;
				} 
				else
				{
					useTextureCoordArray = false;
				}

				m->m_objectArray[id].m_numVertices = vertex_count;
				m->m_objectArray[id].m_vertices = new vertex_t[vertex_count];
				memset(m->m_objectArray[id].m_vertices,0,sizeof(vertex_t) * vertex_count);

				int face_count = scene->objs[i].mesh.faceCount;
				m->m_objectArray[id].m_numFaces = face_count;
				m->m_objectArray[id].m_faces = new unsigned int[face_count * 3];

				for (int j = 0; j < face_count; j++)
				{
					for (int k = 0; k < 3; k++)
					{
						int vid = scene->objs[i].mesh.faces[j].vertex[k];
						int tid = scene->objs[i].mesh.faces[j].texture_coordinates[k];
						int oid = (useTextureCoordArray)? tid:vid;

						m->m_objectArray[id].m_vertices[oid].vertex[0] = scene->objs[i].mesh.vertices[vid].x;
						m->m_objectArray[id].m_vertices[oid].vertex[1] = scene->objs[i].mesh.vertices[vid].z;
						m->m_objectArray[id].m_vertices[oid].vertex[2] = -scene->objs[i].mesh.vertices[vid].y;

						if (scene->objs[i].mesh.texture_coordinates)
						{
							m->m_objectArray[id].m_vertices[oid].texture[0] = scene->objs[i].mesh.texture_coordinates[tid].x;
							m->m_objectArray[id].m_vertices[oid].texture[1] = 1.0f - scene->objs[i].mesh.texture_coordinates[tid].y;
						}
						else
						{
							m->m_objectArray[id].m_vertices[oid].texture[0] = m->m_objectArray[id].m_vertices[oid].texture[1] = 0.0f;
						}

						m->m_objectArray[id].m_faces[j * 3 + k] = oid;
					}
				}

				m->m_objectArray[id].ComputeNormals(IsRightHanded());
			}

			ASE_unloadScene(scene);
		}

		m->Reload();
	}

	return m;
}
// for tetmesh ....
IGraphicsRenderModel* IGraphicsFramework::LoadRenderModel(TetMesh *tetmesh)	
{
	IGraphicsRenderModel* m = CreateRenderModel();

	if (m)
	{
		TissueType *tissue=tetmesh->getTissueData();
		for (int i = 0; i < tetmesh->numTissueTypes(); i++)
		{
			GraphicsRenderObject obj;
			m->m_objectArray.push_back(obj);
			size_t id = m->m_objectArray.size() - 1;
			int vertid=0; int faceid=0;

			// texture
			m->m_objectArray[id].m_tissueType=tissue[i].m_type;
			if (tissue[i].m_type==TetMesh::SKIN)
			{
				m->m_objectArray[id].m_colormap = LoadTexture("tissuext.bmp");
			}
			else if (tissue[i].m_type>=TetMesh::BONE&&tissue[i].m_type<TetMesh::MUSCLES)
			{
				m->m_objectArray[id].m_colormap = LoadTexture("bonetile.bmp");
			}
			else if (tissue[i].m_type>=TetMesh::MUSCLES&&tissue[i].m_type<400)
			{
				m->m_objectArray[id].m_colormap = LoadTexture("musclet.bmp");
			}

			int vertex_count = tissue[i].m_numVertexes;
			m->m_objectArray[id].m_numVertices = vertex_count;
			m->m_objectArray[id].m_vertices = new vertex_t[vertex_count];
			memset(m->m_objectArray[id].m_vertices,0,sizeof(vertex_t) * vertex_count);

			int face_count = tissue[i].m_numFaces;
			m->m_objectArray[id].m_numFaces = face_count;
			m->m_objectArray[id].m_faces = new unsigned int[face_count * 3];

			// set the faces and vertexes
			for(int j=0;j<tetmesh->numTets();j++)
			{
				Tetrahedron *tetrahe=tetmesh->getTetAt(j);
				if(tetrahe->getType()==tissue[i].m_type)
				{
					if(tetrahe->renderFaceAt(0))
					{
						int *vertindex=tetrahe->getVertex();
						// vertex 0
						m->m_objectArray[id].m_vertices[vertid].vertex[0]=tetmesh->getVertexAt(vertindex[0]).getX();
						m->m_objectArray[id].m_vertices[vertid].vertex[1]=tetmesh->getVertexAt(vertindex[0]).getY();
						m->m_objectArray[id].m_vertices[vertid].vertex[2]=tetmesh->getVertexAt(vertindex[0]).getZ();
						m->m_objectArray[id].m_vertices[vertid].normal[0]=tetmesh->getVertexAt(vertindex[0]).getNormalX();
						m->m_objectArray[id].m_vertices[vertid].normal[1]=tetmesh->getVertexAt(vertindex[0]).getNormalY();
						m->m_objectArray[id].m_vertices[vertid].normal[2]=tetmesh->getVertexAt(vertindex[0]).getNormalZ();
						m->m_objectArray[id].m_faces[faceid]=vertid;
						
						m->m_objectArray[id].m_vertices[vertid].texture[0] = 
							m->m_objectArray[id].m_vertices[vertid].texture[1] = 0.0f;
						vertid++;  faceid++;
						// vertex 2
						m->m_objectArray[id].m_vertices[vertid].vertex[0]=tetmesh->getVertexAt(vertindex[2]).getX();
						m->m_objectArray[id].m_vertices[vertid].vertex[1]=tetmesh->getVertexAt(vertindex[2]).getY();
						m->m_objectArray[id].m_vertices[vertid].vertex[2]=tetmesh->getVertexAt(vertindex[2]).getZ();
						m->m_objectArray[id].m_vertices[vertid].normal[0]=tetmesh->getVertexAt(vertindex[2]).getNormalX();
						m->m_objectArray[id].m_vertices[vertid].normal[1]=tetmesh->getVertexAt(vertindex[2]).getNormalY();
						m->m_objectArray[id].m_vertices[vertid].normal[2]=tetmesh->getVertexAt(vertindex[2]).getNormalZ();
						m->m_objectArray[id].m_faces[faceid]=vertid;
						
						m->m_objectArray[id].m_vertices[vertid].texture[0] = 
							m->m_objectArray[id].m_vertices[vertid].texture[1] = 0.0f;
						vertid++; faceid++;
						// vertex 1
						m->m_objectArray[id].m_vertices[vertid].vertex[0]=tetmesh->getVertexAt(vertindex[1]).getX();
						m->m_objectArray[id].m_vertices[vertid].vertex[1]=tetmesh->getVertexAt(vertindex[1]).getY();
						m->m_objectArray[id].m_vertices[vertid].vertex[2]=tetmesh->getVertexAt(vertindex[1]).getZ();
						m->m_objectArray[id].m_vertices[vertid].normal[0]=tetmesh->getVertexAt(vertindex[1]).getNormalX();
						m->m_objectArray[id].m_vertices[vertid].normal[1]=tetmesh->getVertexAt(vertindex[1]).getNormalY();
						m->m_objectArray[id].m_vertices[vertid].normal[2]=tetmesh->getVertexAt(vertindex[1]).getNormalZ();
						m->m_objectArray[id].m_faces[faceid]=vertid;
						
						m->m_objectArray[id].m_vertices[vertid].texture[0] = 
							m->m_objectArray[id].m_vertices[vertid].texture[1] = 0.0f;
						vertid++; faceid++;
					}
					if(tetrahe->renderFaceAt(1))
					{
						int *vertindex=tetrahe->getVertex();
						// vertex 0
						m->m_objectArray[id].m_vertices[vertid].vertex[0]=tetmesh->getVertexAt(vertindex[0]).getX();
						m->m_objectArray[id].m_vertices[vertid].vertex[1]=tetmesh->getVertexAt(vertindex[0]).getY();
						m->m_objectArray[id].m_vertices[vertid].vertex[2]=tetmesh->getVertexAt(vertindex[0]).getZ();
						m->m_objectArray[id].m_vertices[vertid].normal[0]=tetmesh->getVertexAt(vertindex[0]).getNormalX();
						m->m_objectArray[id].m_vertices[vertid].normal[1]=tetmesh->getVertexAt(vertindex[0]).getNormalY();
						m->m_objectArray[id].m_vertices[vertid].normal[2]=tetmesh->getVertexAt(vertindex[0]).getNormalZ();
						m->m_objectArray[id].m_faces[faceid]=vertid;
						
						m->m_objectArray[id].m_vertices[vertid].texture[0] = 
							m->m_objectArray[id].m_vertices[vertid].texture[1] = 0.0f;
						vertid++;  faceid++;
						// vertex 3
						m->m_objectArray[id].m_vertices[vertid].vertex[0]=tetmesh->getVertexAt(vertindex[3]).getX();
						m->m_objectArray[id].m_vertices[vertid].vertex[1]=tetmesh->getVertexAt(vertindex[3]).getY();
						m->m_objectArray[id].m_vertices[vertid].vertex[2]=tetmesh->getVertexAt(vertindex[3]).getZ();
						m->m_objectArray[id].m_vertices[vertid].normal[0]=tetmesh->getVertexAt(vertindex[3]).getNormalX();
						m->m_objectArray[id].m_vertices[vertid].normal[1]=tetmesh->getVertexAt(vertindex[3]).getNormalY();
						m->m_objectArray[id].m_vertices[vertid].normal[2]=tetmesh->getVertexAt(vertindex[3]).getNormalZ();
						m->m_objectArray[id].m_faces[faceid]=vertid;
						
						m->m_objectArray[id].m_vertices[vertid].texture[0] = 
							m->m_objectArray[id].m_vertices[vertid].texture[1] = 0.0f;
						vertid++;  faceid++;
						// vertex 2
						m->m_objectArray[id].m_vertices[vertid].vertex[0]=tetmesh->getVertexAt(vertindex[2]).getX();
						m->m_objectArray[id].m_vertices[vertid].vertex[1]=tetmesh->getVertexAt(vertindex[2]).getY();
						m->m_objectArray[id].m_vertices[vertid].vertex[2]=tetmesh->getVertexAt(vertindex[2]).getZ();
						m->m_objectArray[id].m_vertices[vertid].normal[0]=tetmesh->getVertexAt(vertindex[2]).getNormalX();
						m->m_objectArray[id].m_vertices[vertid].normal[1]=tetmesh->getVertexAt(vertindex[2]).getNormalY();
						m->m_objectArray[id].m_vertices[vertid].normal[2]=tetmesh->getVertexAt(vertindex[2]).getNormalZ();
						m->m_objectArray[id].m_faces[faceid]=vertid;
						
						m->m_objectArray[id].m_vertices[vertid].texture[0] = 
							m->m_objectArray[id].m_vertices[vertid].texture[1] = 0.0f;
						vertid++; faceid++;
					}
					if(tetrahe->renderFaceAt(2))
					{
						int *vertindex=tetrahe->getVertex();
						// vertex 0
						m->m_objectArray[id].m_vertices[vertid].vertex[0]=tetmesh->getVertexAt(vertindex[0]).getX();
						m->m_objectArray[id].m_vertices[vertid].vertex[1]=tetmesh->getVertexAt(vertindex[0]).getY();
						m->m_objectArray[id].m_vertices[vertid].vertex[2]=tetmesh->getVertexAt(vertindex[0]).getZ();
						m->m_objectArray[id].m_vertices[vertid].normal[0]=tetmesh->getVertexAt(vertindex[0]).getNormalX();
						m->m_objectArray[id].m_vertices[vertid].normal[1]=tetmesh->getVertexAt(vertindex[0]).getNormalY();
						m->m_objectArray[id].m_vertices[vertid].normal[2]=tetmesh->getVertexAt(vertindex[0]).getNormalZ();
						m->m_objectArray[id].m_faces[faceid]=vertid;
						
						m->m_objectArray[id].m_vertices[vertid].texture[0] = 
							m->m_objectArray[id].m_vertices[vertid].texture[1] = 0.0f;
						vertid++;  faceid++;
						// vertex 1
						m->m_objectArray[id].m_vertices[vertid].vertex[0]=tetmesh->getVertexAt(vertindex[1]).getX();
						m->m_objectArray[id].m_vertices[vertid].vertex[1]=tetmesh->getVertexAt(vertindex[1]).getY();
						m->m_objectArray[id].m_vertices[vertid].vertex[2]=tetmesh->getVertexAt(vertindex[1]).getZ();
						m->m_objectArray[id].m_vertices[vertid].normal[0]=tetmesh->getVertexAt(vertindex[1]).getNormalX();
						m->m_objectArray[id].m_vertices[vertid].normal[1]=tetmesh->getVertexAt(vertindex[1]).getNormalY();
						m->m_objectArray[id].m_vertices[vertid].normal[2]=tetmesh->getVertexAt(vertindex[1]).getNormalZ();
						m->m_objectArray[id].m_faces[faceid]=vertid;
						
						m->m_objectArray[id].m_vertices[vertid].texture[0] = 
							m->m_objectArray[id].m_vertices[vertid].texture[1] = 0.0f;
						vertid++;  faceid++;
						// vertex 2
						m->m_objectArray[id].m_vertices[vertid].vertex[0]=tetmesh->getVertexAt(vertindex[3]).getX();
						m->m_objectArray[id].m_vertices[vertid].vertex[1]=tetmesh->getVertexAt(vertindex[3]).getY();
						m->m_objectArray[id].m_vertices[vertid].vertex[2]=tetmesh->getVertexAt(vertindex[3]).getZ();
						m->m_objectArray[id].m_vertices[vertid].normal[0]=tetmesh->getVertexAt(vertindex[3]).getNormalX();
						m->m_objectArray[id].m_vertices[vertid].normal[1]=tetmesh->getVertexAt(vertindex[3]).getNormalY();
						m->m_objectArray[id].m_vertices[vertid].normal[2]=tetmesh->getVertexAt(vertindex[3]).getNormalZ();
						m->m_objectArray[id].m_faces[faceid]=vertid;
						
						m->m_objectArray[id].m_vertices[vertid].texture[0] = 
							m->m_objectArray[id].m_vertices[vertid].texture[1] = 0.0f;
						vertid++; faceid++;
					}
					if(tetrahe->renderFaceAt(3))
					{
						int *vertindex=tetrahe->getVertex();
						// vertex 1
						m->m_objectArray[id].m_vertices[vertid].vertex[0]=tetmesh->getVertexAt(vertindex[1]).getX();
						m->m_objectArray[id].m_vertices[vertid].vertex[1]=tetmesh->getVertexAt(vertindex[1]).getY();
						m->m_objectArray[id].m_vertices[vertid].vertex[2]=tetmesh->getVertexAt(vertindex[1]).getZ();
						m->m_objectArray[id].m_vertices[vertid].normal[0]=tetmesh->getVertexAt(vertindex[1]).getNormalX();
						m->m_objectArray[id].m_vertices[vertid].normal[1]=tetmesh->getVertexAt(vertindex[1]).getNormalY();
						m->m_objectArray[id].m_vertices[vertid].normal[2]=tetmesh->getVertexAt(vertindex[1]).getNormalZ();
						m->m_objectArray[id].m_faces[faceid]=vertid;
						
						m->m_objectArray[id].m_vertices[vertid].texture[0] = 
							m->m_objectArray[id].m_vertices[vertid].texture[1] = 0.0f;
						vertid++;  faceid++;
						// vertex 2
						m->m_objectArray[id].m_vertices[vertid].vertex[0]=tetmesh->getVertexAt(vertindex[2]).getX();
						m->m_objectArray[id].m_vertices[vertid].vertex[1]=tetmesh->getVertexAt(vertindex[2]).getY();
						m->m_objectArray[id].m_vertices[vertid].vertex[2]=tetmesh->getVertexAt(vertindex[2]).getZ();
						m->m_objectArray[id].m_vertices[vertid].normal[0]=tetmesh->getVertexAt(vertindex[2]).getNormalX();
						m->m_objectArray[id].m_vertices[vertid].normal[1]=tetmesh->getVertexAt(vertindex[2]).getNormalY();
						m->m_objectArray[id].m_vertices[vertid].normal[2]=tetmesh->getVertexAt(vertindex[2]).getNormalZ();
						m->m_objectArray[id].m_faces[faceid]=vertid;
						
						m->m_objectArray[id].m_vertices[vertid].texture[0] = 
							m->m_objectArray[id].m_vertices[vertid].texture[1] = 0.0f;
						vertid++;  faceid++;
						// vertex 3
						m->m_objectArray[id].m_vertices[vertid].vertex[0]=tetmesh->getVertexAt(vertindex[3]).getX();
						m->m_objectArray[id].m_vertices[vertid].vertex[1]=tetmesh->getVertexAt(vertindex[3]).getY();
						m->m_objectArray[id].m_vertices[vertid].vertex[2]=tetmesh->getVertexAt(vertindex[3]).getZ();
						m->m_objectArray[id].m_vertices[vertid].normal[0]=tetmesh->getVertexAt(vertindex[3]).getNormalX();
						m->m_objectArray[id].m_vertices[vertid].normal[1]=tetmesh->getVertexAt(vertindex[3]).getNormalY();
						m->m_objectArray[id].m_vertices[vertid].normal[2]=tetmesh->getVertexAt(vertindex[3]).getNormalZ();
						m->m_objectArray[id].m_faces[faceid]=vertid;
						
						m->m_objectArray[id].m_vertices[vertid].texture[0] = 
							m->m_objectArray[id].m_vertices[vertid].texture[1] = 0.0f;
						vertid++; faceid++;
					}
				}
			}
		}
		m->Reload();
	}
	return m;
}

IGraphicsRenderModel* IGraphicsFramework::CloneRenderModel(IGraphicsRenderModel* model)								// make a copy of a render model
{
	if (!model) return 0;
	IGraphicsRenderModel* m = CreateRenderModel();

	if (m)
	{
		GraphicsRenderObject obj;
		for (size_t i = 0; i < model->m_objectArray.size(); i++)
		{
			m->m_objectArray.push_back(obj);
			model->m_objectArray[i].Clone(&m->m_objectArray[m->m_objectArray.size() - 1]);
		}

		m->Reload();
	}

	return m;
}