
// File: <graphics_framework.h>
// Copyright (c) 1997-2007, Reachin Technologies AB.
// All rights reserved.
// The copyright to the computer program herein is the property of Reachin
// Technologies AB, Sweden. The program may be used and/or copied only with the
// written permission of Reachin Technologies AB or in accordance with the terms
// stipulated in the agreement under which the program has been supplied.

/** Edited by Rolf Anders Syvertsen
	04.05.2007
	rolfans@stud.ntnu.no
*/

#ifndef _GRAPHICS_FRAMEWORK_H_
#define _GRAPHICS_FRAMEWORK_H_

#include <windows.h>
#include <vector>

#include <HaptX.h>

#include "../tetmesh.h"

#ifndef _WINGDI_
struct BITMAPFILEHEADER
{
	WORD    bfType;
	DWORD   bfSize;
	WORD    bfReserved1;
	WORD    bfReserved2;
	DWORD   bfOffBits;
};

struct BITMAPINFOHEADER
{
	DWORD      biSize;
	LONG       biWidth;
	LONG       biHeight;
	WORD       biPlanes;
	WORD       biBitCount;
	DWORD      biCompression;
	DWORD      biSizeImage;
	LONG       biXPelsPerMeter;
	LONG       biYPelsPerMeter;
	DWORD      biClrUsed;
	DWORD      biClrImportant;
};
#endif

struct image_t
{
	int				width;
	int				height;
	unsigned char*	data;
	void			free_data()	{ if (data) { delete [] data; data = 0; } }
};

struct vertex_t
{
	float vertex[3];
	float normal[3];
	float texture[2];
};

class IGraphicsTexture
{
public:
	IGraphicsTexture()			{}
	virtual ~IGraphicsTexture()	{}

	virtual void				Bind() = 0;							// activate this texture
	virtual void				Unbind() = 0;						// inactivate this texture
	virtual void				Reload() = 0;						// reloads the texture into graphics memory
	virtual	IGraphicsTexture*	Clone(IGraphicsTexture* obj) = 0;	// make a clone of the texture
	virtual image_t				GetImage() = 0;
};


class GraphicsRenderObject
{
public:
	vertex_t*			m_vertices;
	unsigned int*		m_faces;
	int					m_numFaces;
	int					m_numVertices;
	IGraphicsTexture*	m_colormap;
	int					m_tissueType;
	HaptX::iHapticShapeID m_haptx_shapeID;

public:
	GraphicsRenderObject();
	~GraphicsRenderObject();

	void FreeObject();											// free any allocated vertices and faces
	void Clone(GraphicsRenderObject* obj);						// clone the object
	void ComputeNormals(bool right_handed);						// compute vertex normals for the object
};


class IGraphicsRenderModel
{
public:
	std::vector<GraphicsRenderObject> m_objectArray;

public:
	IGraphicsRenderModel()			{}
	virtual ~IGraphicsRenderModel()	{}

	virtual void		Render(bool useMaterials) = 0;									// render the model (textures will be used if useMaterials == true)
	virtual void		Reload() = 0;													// reload the model into graphics memory
	virtual void		SetTransformMatrix(const float* mat) = 0;						// set the models transformation matrix (float[16])
	virtual void		SetTransformMatrix(float pos_x, float pos_y, float pos_z,
										   float rot_x, float rot_y, float rot_z,
										   float scl_x = 1.0f, 
										   float scl_y = 1.0f,
										   float scl_z = 1.0f) = 0;						// set the models transformation matrix (pos = position, rot = rotation in radians, scl = scale)
	
	virtual float*		GetTransformMatrix() = 0;										// retrieve the models transformation matrix (float[16])
	void				ComputeNormals(bool right_handed);								// compute the normals for this model

	virtual void		PushTransform() = 0;											// bind the transformation matrix
	virtual void		PopTransform() = 0;												// unbind th transformation matrix
};


class IGraphicsFramework
{
public:
	IGraphicsFramework()				{}
	virtual ~IGraphicsFramework()		{}

	virtual const char*				GetAPIName() = 0;																	// get the name of the graphics API
	virtual bool					Init(HWND hwnd, int width, int height, bool fullscreen) = 0;						// create a window
	virtual void					Destroy() = 0;																		// destroy the graphics interface and window
	virtual void					Resize(int width, int height) = 0;													// resize the window
	virtual void					SetupPerspectiveProjection(float fov, float aspect, float znear, float zfar) = 0;	// set world projection matrix

	virtual void					SetLighting(bool isEnabled) = 0;													// enable/disable lighting
	virtual void					SetCulling(bool isEnabled) = 0;														// enable/disable culling
	virtual void					SetDepthMask(bool isEnabled) = 0;													// enable/disable depth writes
	virtual void					SetDepthTest(bool isEnabled) = 0;													// enable/disable depth testing

	image_t							LoadImageBMP(const char* strFile);													// load a BMP file type image

	// load a texture (only BMP files are supported)
	virtual IGraphicsTexture*		LoadTexture(const char* strFile) = 0;
	// create an empty render model
	virtual IGraphicsRenderModel*	CreateRenderModel() = 0;
	// load a render model (obky ASE files are supported)
	IGraphicsRenderModel*			LoadRenderModel(const char* strFile);	
	// load a render model from a tetrahedral mesh (.ele file tetmesh class)
	// added by Rolf Anders Syvertsen 07.05.2007
	IGraphicsRenderModel*			LoadRenderModel(TetMesh *tetmesh);
	// make a copy of a render model
	IGraphicsRenderModel*			CloneRenderModel(IGraphicsRenderModel* model);
	
	// is the coordinate system right handed?
	virtual bool					IsRightHanded() = 0;

	virtual void					AddRenderModel(IGraphicsRenderModel* model) = 0;									// add a render model into the renderer
	virtual void					RemoveRenderModel(IGraphicsRenderModel* model) = 0;									// remove a render model from the renderer
	virtual void					RemoveAllRenderModels() = 0;														// remove all render models from the renderer
	virtual void					RenderScene(bool useMaterials) = 0;													// render all render models that have been added into the renderer
	virtual void					RenderLineArray(float* verts, int size) = 0;										// renders lines (must be at least 6 floats in size)
	virtual void					RenderBegin() = 0;																	// called before any render calls are made
	virtual void					RenderFinish() = 0;																	// called after all render calls have been made

	virtual float*					GetCameraMatrix() = 0;																// retrieve the camera matrix (float[16])
	virtual void					CameraGetPosition(float& x, float& y, float& z) = 0;								// retrieve camera position vector
	virtual void					CameraReset() = 0;																	// set camera position to zero
	virtual void					CameraMoveForwardBackward(float val, bool add_y) = 0;								// move camera forward/backward
	virtual void					CameraMoveUpDown(float val) = 0;													// move camera up/down
	virtual void					CameraMoveLeftRight(float val) = 0;													// move camera left/right
	virtual void					CameraRotate(float ang_x, float ang_y) = 0;											// set camera rotation
	virtual void					SetCamera(const float* matrix) = 0;													// set the camera model view matrix directly
	virtual void					SetCamera(float rix, float riy, float riz,
											  float upx, float upy, float upz,
											  float drx, float dry, float drz,
											  float px, float py, float pz) = 0;										// set the model view matrix using right, up, direction and position vectors

	virtual void					SetLightPosition(float x, float y, float z, float w) = 0;							// set light position
	virtual void					SetLightColor(float r, float g, float b, float a) = 0;								// set light color

	virtual void					SetCullBackFace(bool cull_backfaces) = 0;											// should we cull backfaces or frontfaces 

	virtual void					SetColor(float r, float g, float b, float a) = 0;									// set the current active color to be used on models and/or text

	virtual int						GetFontHeight() = 0;																// get font height
	virtual void					PrintText(int x, int y, const char* str, ...) = 0;									// prints out text
};

#endif