
// File: <graphics_opengl.cpp>
// Copyright (c) 1997-2007, Reachin Technologies AB.
// All rights reserved.
// The copyright to the computer program herein is the property of Reachin
// Technologies AB, Sweden. The program may be used and/or copied only with the
// written permission of Reachin Technologies AB or in accordance with the terms
// stipulated in the agreement under which the program has been supplied.

#include "graphics_opengl.h"
#include <math.h>

// ###############################################################################################################

TextureOpenGL::TextureOpenGL(image_t& image)
{
	m_textureID = 0;
	m_imageCopy.width = image.width;
	m_imageCopy.height = image.height;

	if (image.data)
	{
		m_imageCopy.data = new unsigned char[image.width * image.height * 3];
		memcpy(m_imageCopy.data,image.data,image.width * image.height * 3);
	}
	else
	{
		m_imageCopy.data = 0;
	}

	Reload();
}

TextureOpenGL::~TextureOpenGL()
{
	m_imageCopy.free_data();
	glDeleteTextures(1,&m_textureID);
}

void TextureOpenGL::Bind()
{
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,m_textureID);
}

void TextureOpenGL::Unbind()
{
	glDisable(GL_TEXTURE_2D);
}

void TextureOpenGL::Reload()
{
	if (glIsTexture(m_textureID) == GL_TRUE)
		glDeleteTextures(1,&m_textureID);

	glGenTextures(1,&m_textureID);
	glBindTexture(GL_TEXTURE_2D,m_textureID);
	
	if (m_imageCopy.data) gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGB,m_imageCopy.width,m_imageCopy.height,GL_RGB,GL_UNSIGNED_BYTE,m_imageCopy.data);

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
}

IGraphicsTexture* TextureOpenGL::Clone(IGraphicsTexture* obj)
{
	if (!m_imageCopy.data) return 0;
	
	image_t img;
	img.width = ((TextureOpenGL*)obj)->m_imageCopy.width;
	img.height = ((TextureOpenGL*)obj)->m_imageCopy.height;
	img.data = new unsigned char[img.width * img.height * 3];
	memcpy(img.data,((TextureOpenGL*)obj)->m_imageCopy.data,sizeof(unsigned char) * img.width * img.height * 3);

	IGraphicsTexture* tex = new TextureOpenGL(img);
	return tex;
}

image_t TextureOpenGL::GetImage()
{
	return m_imageCopy;
}

// ###############################################################################################################

RenderModelOpenGL::RenderModelOpenGL()
{
	memset(m_transform,0,sizeof(float) * 16);
	m_transform[0] = m_transform[5] = m_transform[10] = m_transform[15] = 1.0f;
}

RenderModelOpenGL::~RenderModelOpenGL()
{
	for (size_t i = 0; i < m_objectArray.size(); i++)
		m_objectArray[i].FreeObject();

	if (!m_objectArray.empty()) m_objectArray.clear();
}

void RenderModelOpenGL::Render(bool useMaterials)
{
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	glPushMatrix();
		glMultMatrixf(m_transform);
		for (size_t i = 0; i < m_objectArray.size(); i++)
		{
			if (!m_objectArray[i].m_vertices) continue;
			if (m_objectArray[i].m_colormap) m_objectArray[i].m_colormap->Bind();

			glVertexPointer(3,GL_FLOAT,sizeof(vertex_t),&m_objectArray[i].m_vertices[0].vertex[0]);
			glNormalPointer(GL_FLOAT,sizeof(vertex_t),&m_objectArray[i].m_vertices[0].normal[0]);
			glTexCoordPointer(2,GL_FLOAT,sizeof(vertex_t),&m_objectArray[i].m_vertices[0].texture[0]);

			glDrawElements(GL_TRIANGLES,m_objectArray[i].m_numFaces * 3,GL_UNSIGNED_INT,m_objectArray[i].m_faces);

			if (m_objectArray[i].m_colormap) m_objectArray[i].m_colormap->Unbind();
		}
	glPopMatrix();

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

void RenderModelOpenGL::Reload()
{
	for (size_t i = 0; i < m_objectArray.size(); i++)
	{
		if (m_objectArray[i].m_colormap) m_objectArray[i].m_colormap->Reload();
	}
}

void RenderModelOpenGL::SetTransformMatrix(const float* mat)
{
	memcpy(m_transform,mat,sizeof(float) * 16);
}

void RenderModelOpenGL::SetTransformMatrix(float pos_x, float pos_y, float pos_z,
										   float rot_x, float rot_y, float rot_z,
										   float scl_x, float scl_y, float scl_z)
{
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(pos_x,pos_y,pos_z);
	glRotatef(rot_x * 57.29578f,1.0f,0.0f,0.0f);
	glRotatef(rot_y * 57.29578f,0.0f,1.0f,0.0f);
	glRotatef(rot_z * 57.29578f,0.0f,0.0f,1.0f);
	glScalef(scl_x,scl_y,scl_z);
	glGetFloatv(GL_MODELVIEW_MATRIX,m_transform);
	glPopMatrix();
}

float* RenderModelOpenGL::GetTransformMatrix()
{
	return m_transform;
}

void RenderModelOpenGL::PushTransform()
{
	glPushMatrix();
	glMultMatrixf(m_transform);
}

void RenderModelOpenGL::PopTransform()
{
	glPopMatrix();
}

// ###############################################################################################################


GraphicsOpenGL::GraphicsOpenGL()
{
	m_fontSize = 10;
	m_hRC = 0;
	m_hDC = 0;

	m_lightPos[0] = m_lightPos[1] = m_lightPos[2] = 0.0f; m_lightPos[3] = 1.0f;
	m_lightCol[0] = m_lightCol[1] = m_lightCol[2] = m_lightCol[3] = 1.0f;
	
	memset(m_cameraMatrix,0,sizeof(float) * 16);
	m_cameraMatrix[0] = m_cameraMatrix[5] = m_cameraMatrix[10] = m_cameraMatrix[15] = 1.0f;
}

GraphicsOpenGL::~GraphicsOpenGL()
{
	Destroy();
}

void GraphicsOpenGL::InitFont()
{
	HFONT font;
	HFONT oldfont;
	m_fontBase = glGenLists(256);

	font = CreateFont(m_fontSize + 4,0,0,0,FW_NORMAL,FALSE,FALSE,FALSE,
					  ANSI_CHARSET,OUT_TT_PRECIS,CLIP_DEFAULT_PRECIS,ANTIALIASED_QUALITY,
					  FF_DONTCARE|DEFAULT_PITCH,"Arial");

	oldfont = (HFONT)SelectObject(m_hDC,font);
	wglUseFontBitmaps(m_hDC,0,255,m_fontBase);
	SelectObject(m_hDC,oldfont);
	DeleteObject(font);
}

void GraphicsOpenGL::DestroyFont()
{
	glDeleteLists(m_fontBase,256);
}

const char* GraphicsOpenGL::GetAPIName()
{
	return "OpenGL";
}

bool GraphicsOpenGL::Init(HWND hwnd, int width, int height, bool fullscreen)
{
	GLuint PixelFormat;
	m_hDC = GetDC(hwnd);

	static PIXELFORMATDESCRIPTOR pfd =
	{
		sizeof(PIXELFORMATDESCRIPTOR),
		1,
		PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,
		32,					// color buffer bits
		0, 0, 0, 0, 0, 0,
		0,
		0,
		0, 0, 0, 0,
		16,					// depth buffer bits
		0,
		0,
		PFD_MAIN_PLANE,
		0,
		0, 0, 0
	};

	if (!(PixelFormat = ChoosePixelFormat(m_hDC,&pfd)))
		return false;

	if (!SetPixelFormat(m_hDC,PixelFormat,&pfd))
		return false;

	if (!(m_hRC = wglCreateContext(m_hDC)))
		return false;

	if (!wglMakeCurrent(m_hDC,m_hRC))
		return false;

	// init opengl stuff
	glColor4f(1.0f,1.0f,1.0f,0.0f);
	glShadeModel(GL_SMOOTH);
	glClearColor(0.3f,0.3f,0.3f,0.0f);
	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_CULL_FACE);
	glEnable(GL_NORMALIZE);

	InitFont();

	m_camRot[0] = m_camRot[1] = 0.0f;
	CameraReset();

	return true;
}

void GraphicsOpenGL::Destroy()
{
	if (m_hRC)
	{
		DestroyFont();

		if (!wglMakeCurrent(NULL,NULL))
			return;

		wglDeleteContext(m_hRC);
		m_hRC = 0;
	}
}

void GraphicsOpenGL::Resize(int width, int height)
{
	glViewport(0,0,width,height);
	m_width = width;
	m_height = height;
}

void GraphicsOpenGL::SetupPerspectiveProjection(float fov, float aspect, float znear, float zfar)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective((GLdouble)fov,(GLdouble)aspect,(GLdouble)znear,(GLdouble)zfar);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void GraphicsOpenGL::SetLighting(bool isEnabled)
{
	if (isEnabled)
	{
		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);
	}
	else
	{
		glDisable(GL_LIGHTING);
		glDisable(GL_LIGHT0);
	}
}

void GraphicsOpenGL::SetCulling(bool isEnabled)
{
	if (isEnabled)
		glEnable(GL_CULL_FACE);
	else glDisable(GL_CULL_FACE);
}

void GraphicsOpenGL::SetDepthMask(bool isEnabled)
{
	glDepthMask((isEnabled)? GL_TRUE : GL_FALSE);
}

void GraphicsOpenGL::SetDepthTest(bool isEnabled)
{
	if (isEnabled)
		glEnable(GL_DEPTH_TEST);
	else glDisable(GL_DEPTH_TEST);
}

IGraphicsTexture* GraphicsOpenGL::LoadTexture(const char* strFile)
{
	image_t img = LoadImageBMP(strFile);
	if (!img.data) return 0;

	IGraphicsTexture* m = new TextureOpenGL(img);
	return m;
}

IGraphicsRenderModel* GraphicsOpenGL::CreateRenderModel()
{
	IGraphicsRenderModel* m = new RenderModelOpenGL();
	return m;
}

bool GraphicsOpenGL::IsRightHanded()
{
	return true;
}

void GraphicsOpenGL::AddRenderModel(IGraphicsRenderModel* model)
{
	if (model) m_modelList.push_back(model);
}

void GraphicsOpenGL::RemoveRenderModel(IGraphicsRenderModel* model)
{
	if (model) m_modelList.remove(model);
}

void GraphicsOpenGL::RemoveAllRenderModels()
{
	if (!m_modelList.empty()) m_modelList.clear();
}

void GraphicsOpenGL::RenderScene(bool useMaterials)
{
	for (std::list<IGraphicsRenderModel*>::iterator w = m_modelList.begin(); w != m_modelList.end(); w++)
	{
		(*w)->Render(useMaterials);
	}
}

void GraphicsOpenGL::RenderLineArray(float* verts, int size)
{
	bool isLighting = (glIsEnabled(GL_LIGHTING) == GL_TRUE);
	if (isLighting) glDisable(GL_LIGHTING);
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3,GL_FLOAT,0,verts);
	glDrawArrays(GL_LINES,0,size);
	glDisableClientState(GL_VERTEX_ARRAY);
	if (isLighting) glEnable(GL_LIGHTING);
}

void GraphicsOpenGL::RenderBegin()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadMatrixf(m_cameraMatrix);

	glLightfv(GL_LIGHT0,GL_POSITION,m_lightPos);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,m_lightCol);
}

void GraphicsOpenGL::RenderFinish()
{
	GLenum glerr = glGetError();
	if (glerr != GL_NO_ERROR)
	{
		switch (glerr)
		{
			case GL_INVALID_ENUM: throw "OpenGL Error: Invalid Enum"; break;
			case GL_INVALID_VALUE: throw "OpenGL Error: Invalid Value"; break;
			case GL_INVALID_OPERATION: throw "OpenGL Error: Invalid Operation"; break;
			case GL_STACK_OVERFLOW: throw "OpenGL Error: Stack Overflow"; break;
			case GL_STACK_UNDERFLOW: throw"OpenGL Error: Stack Underflow"; break;
			case GL_OUT_OF_MEMORY: throw "OpenGL Error: Out Of Memory"; break;
			default: throw "OpenGL Error: Unknown";
		}
	}

	SwapBuffers(m_hDC);
}

void GraphicsOpenGL::UpdateCamera()
{
	glPushMatrix();
	glLoadIdentity();
	glRotatef(m_camRot[0] * 57.29578f,1.0f,0.0f,0.0f);
	glRotatef(m_camRot[1] * 57.29578f,0.0f,1.0f,0.0f);
	glTranslatef(m_camPos[0],m_camPos[1],m_camPos[2]);
	glGetFloatv(GL_MODELVIEW_MATRIX,m_cameraMatrix);
	glPopMatrix();
}

float* GraphicsOpenGL::GetCameraMatrix()
{
	return m_cameraMatrix;
}

void GraphicsOpenGL::CameraGetPosition(float& x, float& y, float& z)
{
	x = m_camPos[0]; y = m_camPos[1]; z = m_camPos[2];
}

void GraphicsOpenGL::CameraReset()
{
	m_camPos[0] = m_camPos[1] = m_camPos[2] = 0.0f;
	UpdateCamera();
}

void GraphicsOpenGL::CameraMoveForwardBackward(float val, bool add_y)
{
	float siny = sinf(m_camRot[1]);
	float cosy = cosf(m_camRot[1]);
	float sinx = sinf(m_camRot[0]);
	float cosx = cosf(m_camRot[0]);
	float addv[3];

	if (add_y)
	{
		addv[0] = -cosx * siny;
		addv[1] = sinx;
		addv[2] = cosx * cosy;
	}
	else
	{
		addv[0] = -siny;
		addv[1] = 0.0f;
		addv[2] = cosy;
	}


	m_camPos[0] += val * addv[0];
	m_camPos[1] += val * addv[1];
	m_camPos[2] += val * addv[2];
	UpdateCamera();
}

void GraphicsOpenGL::CameraMoveUpDown(float val)
{
	m_camPos[1] += val;
	UpdateCamera();
}

void GraphicsOpenGL::CameraMoveLeftRight(float val)
{
	m_camPos[0] += val * sinf(m_camRot[1] + 1.570796f);
	m_camPos[2] += val * -cosf(m_camRot[1] + 1.570796f);
	UpdateCamera();
}

void GraphicsOpenGL::CameraRotate(float ang_x, float ang_y)
{
	m_camRot[0] = ang_x;
	m_camRot[1] = ang_y;
	UpdateCamera();
}

void GraphicsOpenGL::SetCamera(const float* matrix)
{
	memcpy(m_cameraMatrix,matrix,sizeof(float) * 16);
}

void GraphicsOpenGL::SetCamera(float rix, float riy, float riz,
						       float upx, float upy, float upz,
						       float drx, float dry, float drz,
						       float px, float py, float pz)
{
	m_cameraMatrix[ 1] = rix;
	m_cameraMatrix[ 2] = upx;
	m_cameraMatrix[ 3] = drx;
	m_cameraMatrix[ 4] = px;

	m_cameraMatrix[ 5] = riy;
	m_cameraMatrix[ 6] = upy;
	m_cameraMatrix[ 7] = dry;
	m_cameraMatrix[ 8] = py;

	m_cameraMatrix[ 9] = riz;
	m_cameraMatrix[10] = upz;
	m_cameraMatrix[11] = drz;
	m_cameraMatrix[12] = pz;

	m_cameraMatrix[13] = 0.0f;
	m_cameraMatrix[14] = 0.0f;
	m_cameraMatrix[15] = 0.0f;
	m_cameraMatrix[16] = 1.0f;
}

void GraphicsOpenGL::SetLightPosition(float x, float y, float z, float w)
{
	m_lightPos[0] = x;
	m_lightPos[1] = y;
	m_lightPos[2] = z;
	m_lightPos[3] = w;
}

void GraphicsOpenGL::SetLightColor(float r, float g, float b, float a)
{
	m_lightCol[0] = r;
	m_lightCol[1] = g;
	m_lightCol[2] = b;
	m_lightCol[3] = a;
}

void GraphicsOpenGL::SetCullBackFace(bool cull_backfaces)
{
	glCullFace((cull_backfaces)? GL_BACK : GL_FRONT);
}

void GraphicsOpenGL::SetColor(float r, float g, float b, float a)
{
	glColor4f(r,g,b,a);
}

int GraphicsOpenGL::GetFontHeight()
{
	return m_fontSize;
}

void GraphicsOpenGL::PrintText(int x, int y, const char* str, ...)
{
	if (!str) return;

	char buffer[256];
	va_list ap;

	va_start(ap,str);
	vsprintf_s(buffer,str,ap);
	va_end(ap);

	glPushAttrib(GL_LIST_BIT | GL_ENABLE_BIT | GL_CURRENT_BIT | GL_TRANSFORM_BIT);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0,(GLdouble)m_width,0.0,(GLdouble)m_height);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glDisable(GL_LIGHTING);
	glDisable(GL_COLOR_MATERIAL);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);

	glRasterPos2i(x,m_height - y - (m_fontSize - 2));
	glListBase(m_fontBase);
	glCallLists((GLsizei)strlen(buffer),GL_UNSIGNED_BYTE,buffer);
	glListBase(0);
	glRasterPos2i(0,0);

	glPopAttrib();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}